class Shard(object):
    def __init__(self, shard_id, node, state, no_of_docs=0):
        self.__shard_id = shard_id
        self.__node = node
        self.__state = state
        self.__no_of_docs = no_of_docs

    @property
    def shard_id(self):
        return self.__shard_id

    @property
    def node(self):
        return self.__node

    @property
    def state(self):
        return self.__state

    @property
    def no_of_docs(self):
        return self.__no_of_docs

    @no_of_docs.setter
    def no_of_docs(self, value):
        self.__no_of_docs = value



class Node(object):

    def __init__(self, node_id, publish_address,
                 node_type=('master', 'data', 'ingest')):
        """
        Representation of an elasticsearch node in the cluster
        :param str node_id: The id of the node
        :param str publish_address: The exposed address of the node
        :param tuple(str) node_type: The type of the node: master data, ingest
        """
        self.__node_id = node_id
        self.__publish_address = publish_address
        self.__node_type = node_type

    @property
    def node_id(self):
        return self.__node_id

    @property
    def publish_address(self):
        return self.__publish_address

    @property
    def node_type(self):
        return self.__node_type


class NodeRegistry(object):
    def __init__(self):
        self.__nodes = {}

    @property
    def nodes(self):
        return self.__nodes

    def get_nodes_from_elastic(self, elk_client):
        """

        :param elasticsearch.Elasticsearch elk_client: Instance of an
        Elasticsearch client to connect to
        """
        node_client = elk_client.nodes
        resp = node_client.info()
        for node_id, node_info in resp['nodes'].items():
            publish_address = node_info['http']['publish_address']
            roles = node_info['roles']
            self.__nodes[node_id] = Node(node_id=node_id,
                                         publish_address=publish_address,
                                         node_type=tuple(roles))

    def get_node_by_id(self, node_id):
        """
        Get node object based on the nodes id
        :param str node_id: The id of the node
        :return: The node object
        :rtype: Node
        """

        return self.__nodes[node_id]



from typing import Any

import numpy


def make_sequence(elements):
    if isinstance(elements, (list, set, tuple)):
        return elements
    elif elements is None:
        return []
    else:
        return [elements]


def sanitize_data(data: Any) -> Any:  # pylint: disable=invalid-name,too-many-return-statements
    """
    Sanitize turns Numpy types into basic Python types so they
    can be serialized into JSON.
    """
    if isinstance(data, (str, float, int, bool)):
        # x is already serializable
        return data
    elif isinstance(data, numpy.ndarray):
        # array needs to be converted to a list
        return data.tolist()
    elif isinstance(data, numpy.number):  # pylint: disable=no-member
        # NumPy numbers need to be converted to Python numbers
        return data.item()
    elif isinstance(data, dict):
        # Dicts need their values sanitized
        return {key: sanitize_data(value) for key, value in data.items()}
    elif isinstance(data, (list, tuple)):
        # Lists and Tuples need their values sanitized
        return [sanitize_data(x_i) for x_i in data]
    elif hasattr(data, 'to_json'):
        return data.to_json()
    return data


import numpy as np
from dask.dataframe.utils import make_meta

from dask_elk.elk_entities.shards import Shard


class IndexNotFoundException(Exception):
    pass


class Index(object):
    def __init__(self, name, shards=None, mapping=None):
        """

        :param str name: The index name
        :param shards: Shards where this index is located in
        :param dict[str, np.dtype] mapping: The dataframe meta info of the
        index
        """
        self.__name = name
        self.__shards = shards
        if not shards:
            self.__shards = []
        self.__shards_dict = {}
        self.__mapping = mapping

    @property
    def name(self):
        return self.__name

    @property
    def shards(self):
        return self.__shards

    @property
    def mapping(self):
        return self.__mapping

    def add_shard(self, shard):
        """
        Add a shard object to the index
        :param dask_elk.elk_entities.shards.Shard shard: The shard to be added
        """
        self.__shards.append(shard)
        self.__shards_dict[shard.shard_id] = shard

    def get_shard_by_id(self, shard_id):
        """

        :param int shard_id: The id of the shard
        """

        return self.__shards_dict[shard_id]


class IndexRegistry(object):

    __MINIMAL_6_x_SUPPORTED_VERSION = ("6", "8", "0")

    def __init__(self, nodes_registry):
        """

        :param dask_elk.elk_entities.nodes.NodeRegistry node_registry: The
        registry containing info on the clusters nodes
        """

        self.__indices = {}
        self.__nodes_registry = nodes_registry

    @property
    def indices(self):
        return self.__indices

    @indices.setter
    def indices(self, value):
        self.__indices = value

    def calculate_meta(self):
        """
        Since Elasticsearch is schemaless it is possible that to indices might
        have different mappings for the same "type" of documents. To get by
        this, merge all meta's together. During reading the Parsers will be
        responsible for handling mssing or different type of data.
        :return: Empty dataframe containing the expected schema
        :rtype: pandas.DataFrame
        """
        meta = {}
        for index in self.indices.values():
            meta.update(index.mapping)

        meta_df = make_meta(meta)
        return meta_df

    def get_indices_from_elasticsearch(self, elk_client, index=None, doc_type="_doc"):
        """
        Get a list of index objects with its respective mappings
        :param elasticsearch.Elasticsearch elk_client: The Elasticsearch client
        instance
        :param str index: String of comma seperated indices
        :param str doc_type: The doc type of the indices
        """
        cluster_info = elk_client.info()
        es_version = tuple(cluster_info["version"]["number"].split("."))
        self.__indices = self.__get_mappings(
            elk_client.indices,
            index=index,
            doc_type=doc_type,
            backward_compatibility=es_version >= self.__MINIMAL_6_x_SUPPORTED_VERSION,
        )

        self.__get_shards_with_nodes(elk_client, index=index)

    def __get_mappings(
        self, index_client, index, doc_type="_doc", backward_compatibility: bool = False
    ):

        try:
            resp = index_client.get_mapping(
                index=index,
                doc_type=doc_type,
                ignore_unavailable=True,
                include_type_name=backward_compatibility,
            )

            indices_dict = {}
            for index_key, mappings in resp.items():
                mapping = mappings["mappings"][doc_type]["properties"]
                mapping = self.__create_mappings_dtypes(mapping)

                index_obj = Index(name=index_key, mapping=mapping)
                indices_dict[index_key] = index_obj
            return indices_dict
        except Exception:
            raise IndexNotFoundException

    @staticmethod
    def __create_mappings_dtypes(mappings):
        """
        Create dictionary with dtypes for each index field. Not supporting
        nested fields yet.
        :param dict[str, dict[str, T]] mappings:
        :return: dict[str, np.dtype]
        """

        index_maps = {}
        for field_name, field_type in mappings.items():
            pandas_type = np.dtype(object)
            type = field_type.get("type")

            # Treat all numbers as floats to allow None values
            if type == "integer" or type == "long" or type == "float":
                pandas_type = np.dtype("float64")
            elif type == "date":
                pandas_type = np.dtype("datetime64[ns]")
            index_maps[field_name] = pandas_type

        index_maps["_id"] = np.dtype(object)

        return index_maps

    def __get_shards_with_nodes(self, elk_client, index=None):

        shard_info = elk_client.search_shards(index=index)
        for shard in shard_info["shards"]:
            for shard_item in shard:
                index_obj = self.__indices[shard_item["index"]]
                shard_id = shard_item["shard"]
                node_id = shard_item["node"]
                state = shard_item["state"]
                primary = shard_item["primary"]
                node = self.__nodes_registry.get_node_by_id(node_id)
                if state == "STARTED" and primary:
                    shard = Shard(shard_id=shard_id, node=node, state=state)
                    index_obj.add_shard(shard)

    @staticmethod
    def get_documents_count(elk_client, query, index, doc_type="_doc", shard=None):
        """
        Method to get the document count for a specified query on a specified
        index. If shard is provided then the query is executed on that
        specific shard
        :param elasticsearch.Elasticsearch elk_client: Elasticsearch client to
        use
        :param dict[str, T] query: Query to push down to elasticsearch
        :param dask_elk.elk_entities.index.Index index: The index object to
        execute query on
        :param str doc_type: The doc type the index belongs to
        :param dask_elk.elk_entities.shards.Shard | None shard: The shard to
        execute the query on
        :return: The number of documents
        :rtype: int
        """
        preference = None
        if shard:
            preference = "_shards:{}".format(shard.shard_id)

        count_argurments = {"body": query, "index": index.name, "doc_type": doc_type}

        if preference:
            count_argurments.update({"preference": preference})

        no_of_documents = elk_client.count(**count_argurments)

        return no_of_documents["count"]



import pandas as pd

from dask import delayed
from elasticsearch import Elasticsearch
from elasticsearch.helpers import scan

from dask_elk.parsers import DocumentParser


class PartitionReader(object):
    def __init__(self, index, shard, meta, node=None, doc_type='_doc',
                 elastic_class=Elasticsearch, query=None,
                 slice_id=None, slice_max=1, scroll_size=1000,
                 client_args=None, **kwargs):
        """
        Intantiate the PartitionReader object. PartitionReader contains the
        the logic to get data from each shard (partition)
        :param dask_elk.elk_entities.index.Index index: The index object
        :param dask_elk.elk_entities.shard.Shard shard: The shard from which to
        read data
        :param dask_elk.elk_entities.node.Node node: The node to fetch data
        from.
        :param pandas.DataFrame meta: Meta info for schema of returned data.
        :param str doc_type: The doc type the index belongs to
        :param elastic_class: Elasticsearch class to create client
        :param dict[str, T] query: The query to be pushed down to the ELK
        :param int id: The scroll id for concurrent scroll query. See
        https://www.elastic.co/guide/en/elasticsearch/reference/current/
        search-request-scroll.html#sliced-scroll
        :param max: Max number of slices
        :param int scroll_size: Number of documents to fetch for eacch scroll
        request
        :param dict[str, T] | None client_args: Additional arguments to pass to
        client.
        :param dict[str, T] kwargs: Additional scan arguments to be passed to
        python's Elasticsearch search method.
        """

        self.__elastic_class = elastic_class
        self.__query = query
        self.__index = index
        self.__shard = shard
        self.__node = node
        self.__meta = meta
        self.__doc_type = doc_type
        self.__id = slice_id
        self.__max = slice_max
        self.__scroll_size = scroll_size
        self.__client_args = client_args
        self.__additional_scan_args = kwargs

    @delayed
    def read(self):
        client_args = {}
        if self.__client_args:
            client_args = self.__client_args.copy()
        client_args['hosts'] = [
                self.__node.publish_address if self.__node else
                self.__shard.node.publish_address]

        client = self.__elastic_class(**client_args)

        if not self.__query:
            self.__query = {}

        scan_args = dict(index=self.__index.name,
                         doc_type=self.__doc_type,
                         preference='_shards:{}'.format(self.__shard.shard_id),
                         query=dict(self.__query), size=self.__scroll_size
                         )

        if self.__max > 1:
            # We need sliced scroll request
            scroll_slice = {'id': self.__id, 'max': self.__max}
            scan_args.get('query', {}).update(
                {'slice': scroll_slice})

        if self.__additional_scan_args:
            scan_args.update(self.__additional_scan_args)

        results = []
        for hit in scan(client, **scan_args):
            result = hit['_source']
            result['_id'] = hit['_id']

            results.append(result)
        data_frame = self.__meta
        if results:
            data_frame = pd.DataFrame(results)
            data_frame = DocumentParser.parse_documents(data_frame,
                                                        self.__meta)
            data_frame = data_frame[self.__meta.columns]

        return data_frame



from elasticsearch import Elasticsearch
import dask.dataframe as dd

from dask_elk.elk_entities.index import IndexRegistry
from dask_elk.elk_entities.node import Node, NodeRegistry
from dask_elk.helpers import make_sequence
from dask_elk.reader import PartitionReader
from .delayed_methods import bulk_save


class DaskElasticClient(object):
    """
    Basic object. Client to connect with Elasticsearch

    :param list[str]|str host: The host to connect to
    :param int port: The port to connect to
    :param elasticsearch.Elasticsearch client_klass: The class to use to
            create the Elasticsearch client instance
    :param str|None username: The username if X-pack is activated
    :param str|None password: The password if X-pack is activated
    :param bool wan_only: If client should perform Node lookup. Set to True
            if ELK behind firewall or if nodes not accessed directly
    :param dict[str,T] client_kwargs: Arguments to pass to Elasticsearch
            client object created by client_klass

    """

    def __init__(self, host='localhost', port=9200, client_klass=Elasticsearch,
                 username=None, password=None, wan_only=False,
                 **client_kwargs):
        """
        Constructor for DaskElasticClient object

        :param list[str] | str host: The host to connect to
        :param int port: The port to connect to
        :param elasticsearch.Elasticsearch client_klass: The class to use to
                create the Elasticsearch client instance
        :param str | None username: The username if X-pack is activated
        :param str | None password: The password if X-pack is activated
        :param bool wan_only: If client should perform Node lookup. Set to True
                if ELK behind firewall or if nodes not accessed directly
        :param dict[str, T] client_kwargs: Arguments to pass to Elasticsearch
                client object created by client_klass
        """
        self.__host = host
        self.__port = port
        self.__client_klass = client_klass
        self.__username = username
        self.__password = password
        self.__wan_only = wan_only
        self.__client_args = self.__create_client_args(client_kwargs)

    @property
    def client_args(self):
        return self.__client_args

    @property
    def hosts(self):
        return self.__host

    @property
    def port(self):
        return self.__port

    @property
    def username(self):
        return self.__username

    @property
    def password(self):
        return self.__password

    @property
    def wan_only(self):
        return self.__wan_only

    def read(self, query=None, index=None, doc_type=None,
             number_of_docs_per_partition=1000000, size=1000,
             fields_as_list=None,
             **kwargs):

        """
        Method to read from elasticsearch index/ices

        :param dict(str,T)|None query: The query to push down to ELK.
        :param str index: String of the index/ices to execute query on.
        :param str doc_type: Type index belongs too
        :param int number_of_docs_per_partition: Number of documents for
            each partition/task created by the readers
        :param int size: The scroll size
        :param str|None fields_as_list: Comma separated list of fields to be
            treated as object
        :param kwargs: Additional keyword arguments to pass to the search
            method of python Elasticsearch client
        :return: Dask Dataframe containing the data
        :rtype: dask.dataframe.DataFrame

        """

        fields_as_list = fields_as_list
        client_cls = self.__client_klass

        elk_client = client_cls(**self.client_args)

        if not query:
            query = {}

        # Get nodes info first
        node_registry = NodeRegistry()
        node_registry.get_nodes_from_elastic(elk_client)

        index_registry = IndexRegistry(nodes_registry=node_registry)
        index_registry.get_indices_from_elasticsearch(elk_client,
                                                      index=index,
                                                      doc_type=doc_type)

        meta = index_registry.calculate_meta()
        if fields_as_list:
            for field in map(str.strip, fields_as_list.split(',')):
                if field in meta.columns:
                    meta[field] = meta[field].astype(object)

        delayed_objs = []
        for index in index_registry.indices.values():
            for shard in index.shards:
                no_of_docs = IndexRegistry.get_documents_count(elk_client,
                                                               query, index,
                                                               shard=shard)
                node = None
                if self.wan_only:
                    node = Node(node_id='master', publish_address=self.hosts)

                number_of_partitions = self.__get_number_of_partitions(
                    no_of_docs, number_of_docs_per_partition)
                for slice_id in range(number_of_partitions):
                    part_reader = self.__create_partition_reader(
                        index,
                        shard,
                        node, query,
                        doc_type,
                        meta,
                        number_of_partitions,
                        slice_id,
                        size,
                        **kwargs
                    )

                    delayed_objs.append(part_reader.read())

        return dd.from_delayed(delayed_objs, meta=meta)

    def save(self, data, index, doc_type, action='index'):
        """
        Save a dask dataframe to Elasticsearch

        :param dask.dataframe.DataFrame data: Dataframe to save into ELK
        :param str index: The index to save dataframe
        :param str doc_type: Index doc type
        :param str action: index if indexing you data 'update' if updating data
        :return: The data with the save applied on it
        :rtype: dask.dataframe.DataFrame

        """
        client_cls = self.__client_klass
        client_args = self.__client_args

        bulk_arguments = {'index': index, 'doc_type': doc_type,
                          'action': action,
                          }
        data = data.map_partitions(bulk_save, client_cls, client_args,
                                   meta=data, **bulk_arguments)

        return data

    def __create_partition_reader(self, index, shard, node, query, doc_type,
                                  meta, number_of_partitions, slice_id, size,
                                  **scan_arguments):
        """
        Create partition reader object
        :param dask_elk.elk_entities.index.Index index: Index to fetch data
        :param dask_elk.elk_entities.shard.Shard shard:
        :param dask_elk.elk_entities.node.Node node:
        :param dict[str, T] query:
        :param str doc_type:
        :param pandas.DataFrame meta:
        :param int number_of_partitions:
        :param int slice_id:
        :param int size: The scroll size
        :return: The partition reader object to read data from
        :rtype: dask_elk.reader.PartitionReader
        """

        part_reader = PartitionReader(
            index=index, shard=shard,
            meta=meta,
            node=node,
            doc_type=doc_type,
            query=query,
            scroll_size=size,
            slice_id=slice_id,
            slice_max=number_of_partitions,
            client_args=self.__client_args,
            **scan_arguments
        )

        if number_of_partitions > 1:
            part_reader = PartitionReader(
                index=index, shard=shard,
                meta=meta,
                node=node,
                doc_type=doc_type,
                query=query,
                scroll_size=size,
                slice_id=slice_id,
                slice_max=number_of_partitions,
                client_args=self.client_args,
                **scan_arguments
            )

        return part_reader

    def __get_number_of_partitions(self, no_of_docs, no_of_docs_per_partition):
        partitions = max(1, int(no_of_docs / no_of_docs_per_partition))
        return partitions

    def __create_client_args(self, client_kwargs):
        client_arguments = {'hosts': make_sequence(self.hosts),
                            'port': self.port}
        if self.username and self.password:
            client_arguments.update(
                {'http_auth': (self.username, self.password)})
        if client_kwargs:
            client_arguments.update(client_kwargs)
        return client_arguments



import numpy as np
import pandas as pd
from elasticsearch.helpers import bulk

from dask_elk import helpers


def bulk_save(partition, client_cls, client_args, **kwargs):
    """
    Method to save a dataframe to Elasticsearch index. This method should be
    applied to all dask dataframe partitions using map_partitions method.

    :param pandas.DataFrame partition: The dataframe partition to save in ELK
    :param type client_cls: Elasticsearch client class to be instantiated
    :param dict[str,T] kwargs: Arguments to be used for creating the bulk
        payload/objects
    :return: The partition dataframe
    :rtype: pd.DataFrame

    """

    elk_client = client_cls(**(client_args or {}))
    records = partition.to_dict(orient='records')
    actions = []

    for record in records:
        record = helpers.sanitize_data(record)
        index = kwargs.get('index').format(**record)
        doc_type = kwargs.get('doc_type').format(**record)
        op_type = kwargs['action']

        action = dict(_index=index, _type=doc_type, _op_type=op_type)
        if op_type == 'update':
            action['_id'] = record['_id']
            del record['_id']
            action.update({'doc': record})
        else:
            action.update(record)
        actions.append(action)

    bulk(elk_client, actions, stats_only=False)

    return partition



class DocumentParser(object):

    @classmethod
    def parse_documents(self, documents, meta):
        """

        :param pandas.DataFrame documents:
        :param pandas.DataFrame meta:
        :return: the document formated
        """
        meta_columns = meta.columns
        documents_columns = documents.columns

        missing_columns = self.find_missing_columns(meta,
                                                    documents)
        extra_columns = [column for column in documents_columns if
                         column not in meta_columns]

        for column_name, dtype in missing_columns.items():
            documents[column_name] = None
            documents[column_name] = documents[column_name].astype(dtype)

        columns_to_cast = [column_name for column_name in meta_columns if
                           column_name not in missing_columns]

        for column_name in columns_to_cast:
            documents[column_name] = documents[column_name].astype(
                meta.dtypes[column_name])

        if extra_columns:
            documents = documents.drop(labels=extra_columns, axis=1)

        return documents

    @staticmethod
    def find_missing_columns(meta, documents):
        """
        :param pandas.DataFrame meta:
        :param pandas.DataFrame documents:
        :return:
        :rtype: dict[str, np.dtypes]
        """
        meta_columns = meta.columns
        documents_columns = documents.columns

        diff = [column_name for column_name in meta_columns if
                column_name not in documents_columns]

        missing_columns = {}
        for column_name in diff:
            missing_columns[column_name] = meta.dtypes[column_name]

        return missing_columns



from setuptools import setup

long_description = ''

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='dask_elk',
    version='0.4.0',
    description='Dask connection with Elasticsearch',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/avlahop/dask-elk',
    maintainer='Apostolos Vlachopoulos',
    maintainer_email='avlahop@gmail.com',
    keywords='elasticsearch dask parallel',
    licence='GPLv3',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License (GPL)',
    ],
    packages=[
        'dask_elk',
        'dask_elk.elk_entities',
    ],
    python_requires=">=2.7",
    install_requires=[
        'dask[complete]',
        'elasticsearch>=6.2',
        'pandas>=0.19.0',
    ],
    tests_require=[
        'mock',
        'nose',
    ],
    test_suite='nose.collector',
    project_urls = {
        'Documentation': 'https://dask-elk.readthedocs.io'
    }
)